function each(list, decorator) {
    if (list.head === null) {
        return '';
    }

    if (list.head === list.tail) {
        return translate(list.head.data, decorator);
    }

    return list.map(function(node) {
        return translate(node, decorator);
    }).join('');
}

function eachDelim(list, delimeter, decorator) {
    if (list.head === null) {
        return '';
    }

    if (list.head === list.tail) {
        return translate(list.head.data, decorator);
    }

    return list.map(function(node) {
        return translate(node, decorator);
    }).join(delimeter);
}

function translate(node, decorator) {
    var result;

    switch (node.type) {
        case 'StyleSheet':
            result = each(node.rules, decorator);
            break;

        case 'Atrule':
            var nodes = ['@', node.name];

            if (node.expression && !node.expression.sequence.isEmpty()) {
                nodes.push(' ', translate(node.expression));
            }

            if (node.block) {
                nodes.push('{', translate(node.block), '}');
            } else {
                nodes.push(';');
            }

            result = nodes.join('');
            break;

        case 'Ruleset':
            result = translate(node.selector) + '{' + translate(node.block) + '}';
            break;

        case 'Selector':
            result = eachDelim(node.selectors, ',', decorator);
            break;

        case 'SimpleSelector':
            var nodes = node.sequence.map(function(node) {
                // add extra spaces around /deep/ combinator since comment beginning/ending may to be produced
                if (node.type === 'Combinator' && node.name === '/deep/') {
                    return ' ' + translate(node) + ' ';
                }

                return translate(node);
            });

            result = nodes.join('');
            break;

        case 'Block':
            result = eachDelim(node.declarations, ';', decorator);
            break;

        case 'Declaration':
            result = translate(node.property) + ':' + translate(node.value);
            break;

        case 'Property':
            result = node.name;
            break;

        case 'Value':
            result = node.important
                ? each(node.sequence) + '!important'
                : each(node.sequence);
            break;

        case 'Attribute':
            var result = translate(node.name);
            var flagsPrefix = ' ';

            if (node.operator !== null) {
                result += node.operator;

                if (node.value !== null) {
                    result += translate(node.value);

                    // space between string and flags is not required
                    if (node.value.type === 'String') {
                        flagsPrefix = '';
                    }
                }
            }

            if (node.flags !== null) {
                result += flagsPrefix + node.flags;
            }

            result = '[' + result + ']';
            break;

        case 'FunctionalPseudo':
            result = ':' + node.name + '(' + eachDelim(node.arguments, ',') + ')';
            break;

        case 'Function':
            result = node.name + '(' + eachDelim(node.arguments, ',') + ')';
            break;

        case 'Negation':
            result = ':not(' + eachDelim(node.sequence, ',') + ')';
            break;

        case 'Braces':
            result = node.open + each(node.sequence) + node.close;
            break;

        case 'Argument':
        case 'AtruleExpression':
            result = each(node.sequence, decorator);
            break;

        case 'Url':
            result = 'url(' + translate(node.value) + ')';
            break;

        case 'Progid':
            result = translate(node.value);
            break;

        case 'Combinator':
            result = node.name;
            break;

        case 'Identifier':
            result = node.name;
            break;

        case 'UnicodeRange':
            result = node.name;
            break;

        case 'PseudoClass':
            result = ':' + node.name;
            break;

        case 'PseudoElement':
            result = '::' + node.name;
            break;

        case 'Class':
            result = '.' + node.name;
            break;

        case 'Id':
            result = '#' + node.name;
            break;

        case 'Hash':
            result = '#' + node.value;
            break;

        case 'Dimension':
            result = node.value + node.unit;
            break;

        case 'Nth':
            result = node.value;
            break;

        case 'Number':
            result = node.value;
            break;

        case 'String':
            result = node.value;
            break;

        case 'Operator':
            result = node.value;
            break;

        case 'Raw':
            result = node.value;
            break;

        case 'Unknown':
            result = node.value;
            break;

        case 'Percentage':
            result = node.value + '%';
            break;

        case 'Space':
            result = ' ';
            break;

        case 'Comment':
            result = '/*' + node.value + '*/';
            break;

        default:
            throw new Error('Unknown node type: ' + node.type);
    }

    return decorator(result, node);
}

module.exports = translate;
